/* genscan2underlays.c -- convert Genscan output to underlay format
 *
 * The Genscan output file should have the following form:
 *
 *    GENSCAN 1.0     Date run: 31-Aug-99     Time: 07:40:46
 *
 *    Sequence addX : 100000 bp : 43.74% C+G : Isochore 2 (43.00 - 51.00 C+G%)
 *
 *    Parameter matrix: HumanIso.smat
 *
 *    Predicted genes/exons:
 *
 *    Gn.Ex Type S .Begin ...End .Len Fr Ph I/Ac Do/T CodRg P.... Tscr..
 *    ----- ---- - ------ ------ ---- -- -- ---- ---- ----- ----- ------
 *
 *     1.01 Intr +   1470   1515   46  1  1   87   97    27 0.584   0.97
 *     1.02 Intr +   4316   4370   55  2  1  103   89    47 0.802   5.28
 *       etc.
 *
 * This program was suggested by Ken Dewar.
 * The Laj option was added at the suggestion of Mike Wilson.
 *
 * $Revision: 1.19 $$Date: 2001/06/26 23:51:03 $
 */

#include "util.h"

#define SYNTAX  "genscan_file [-laj] [-split] " \
                  "[-fprom fwd_prom_color] [-rprom rev_prom_color] " \
                  "[-fexon fwd_exon_color] [-rexon rev_exon_color] " \
                  "[-fplya fwd_plya_color] [-rplya rev_plya_color] " \
                  "> output_file"

enum { BUFSIZE = 1000 };

static char *Types[6] = {
	"fwd_promoter", "rev_promoter", "fwd_exon", "rev_exon",
	"fwd_polyA", "rev_polyA"
};

static void get_args (int argc, char **argv, char **filename, bool *laj,
                      bool *split, char **colors);
static bool allnull (char **colors, int len);
static void print_underlay (int begin, int end, char *name, char *dir,
                            float prob, bool laj, bool split, char **colors);

int main (int argc, char **argv) {
	char *filename = NULL;                  /* name of genscan file         */
	bool laj = 0;                           /* make labeled underlays?      */
	bool split = 0;                         /* split strands?               */
	char *colors[6] = {NULL, NULL, NULL,    /* color names, in SYNTAX order */
	                   NULL, NULL, NULL};
	FILE *fp;                               /* ptr to input file            */
	char *stat;                             /* read status                  */
	char buf [BUFSIZE];                     /* input buffer                 */
	char pred [BUFSIZE];                    /* prediction type field        */
	char dir [BUFSIZE];                     /* direction field              */
	int addr1, addr2;                       /* address fields               */
	float prob;                             /* probability field            */
	int i;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &filename, &laj, &split, colors);
	if (!filename)
		fatal (usage (SYNTAX));
	if (allnull (colors, 6))
		fatal ("You must specify one or more colors to get output.");

	/* print header */
	printf ("# Underlays converted from \"%s\" by %s.\n\n", filename, argv0);

	/* print color legend */
	for (i = 0; i < 6; i++) {
		if (colors[i])
			printf ("%s %s\n", colors[i], Types[i]);
	}
	putchar ('\n');

	/* open input file; skip intro */
	if (!(fp = fopen (filename, "r")))
		fatalf ("Can't open file \"%s\".", filename);
	while ((stat = fgets (buf, BUFSIZE, fp)) && !strstr (buf, "---"))
		;
	if (!stat)
		fatalf ("Can't find line with \"---\".");

	/* process lines */
	while (fgets (buf, BUFSIZE, fp) && !strstr (buf, "Predicted")) {
		if (*buf == '\n')
			continue;
		if (sscanf (buf+6, "%s %1[+-] %d %d", pred, dir, &addr1, &addr2) != 4)
			fatalf ("Can't parse line:\n%s", buf);
		if (!strsame (pred, "Prom") && !strsame (pred, "Init") &&
				!strsame (pred, "Intr") && !strsame (pred, "Term") &&
				!strsame (pred, "Sngl") && !strsame (pred, "PlyA"))
			fatalf ("Invalid prediction type:\n%s", buf);
		if (addr1 > addr2) {
			swap (&addr1, &addr2);
		}
		prob = -1.0;
		if (!strsame (pred, "Prom") && !strsame (pred, "PlyA")) {   /* exon */
			if (!isdigit (buf[54]) || sscanf (buf+54, "%f", &prob) != 1 ||
					prob < 0.0 || prob > 1.0)
				fatalf ("Invalid exon probability:\n%s", buf);
		}
		print_underlay (addr1, addr2, pred, dir, prob, laj, split, colors);
	}

	/* clean up */
	fclose (fp);
	return 0;
}

static void get_args (int argc, char **argv, char **filename, bool *laj,
                      bool *split, char **colors) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-laj"))
			*laj = 1;
		else if (strsame (argv[i], "-split"))
			*split = 1;
		else if (strsame (argv[i], "-fprom"))
			colors[0] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-rprom"))
			colors[1] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-fexon"))
			colors[2] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-rexon"))
			colors[3] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-fplya"))
			colors[4] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-rplya"))
			colors[5] = optval (argc, argv, &i, SYNTAX);
		else if (!*filename)
			*filename = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

static bool allnull (char **colors, int len) {
	int i;
	for (i = 0; i < len; i++) {
		if (colors[i])
			return 0;
	}
	return 1;
}

static void print_underlay (int begin, int end, char *name, char *dir,
                            float prob, bool laj, bool split, char **colors) {
	int type;
	char *half;

	if (strsame (name, "Prom"))
		type = (*dir == '+' ? 0 : 1);
	else if (strsame (name, "PlyA"))
		type = (*dir == '+' ? 4 : 5);
	else   /* exon */
		type = (*dir == '+' ? 2 : 3);
	if (!colors[type])
		return;    /* skip if no color specified */

	half = (split ? dir : "");

	if (laj) {
		if (strsame (name, "Prom") || strsame (name, "PlyA"))
			printf ("%d %d (Genscan %s %s) %s %s\n",
			        begin, end, name, dir, Types[type], half);
		else
			printf ("%d %d (Genscan %s %s P=%.3f) %s %s\n",
			        begin, end, name, dir, prob, Types[type], half);
	}
	else
		printf ("%d %d %s %s\n", begin, end, Types[type], half);
}
