/* strong_hits.c -- find gap-free matches in an alignment that meet specified
 *                  constraints for length and strength
 *
 * This program examines an lav or concise alignment file to find all
 * gap-free segments meeting user-specified thresholds for length and
 * percent identity.  If an optional file of intervals is provided,
 * the program will also report which of these intervals, if any,
 * overlap each such segment.  Several formats for the interval file
 * are supported, including those for exons, repeats, annotations,
 * and underlays, or any other format having lines containing two
 * positions separated by blanks.  If the file is in exons format, the
 * "-exons-only" option can be used to instruct the program to ignore
 * the intervals for genes and translated regions, and look only for
 * overlaps with exons.  By default the program will look for intervals
 * that overlap the segment's first sequence, but this can be changed
 * to the second sequence by using the "-inseq2" option.
 *
 * This program was suggested by John Pedan.
 *
 * $Revision: 1.7 $$Date: 2001/09/18 00:15:49 $
 */

#include "lav.h"
#include "exons.h"
#include "feat.h"
#include "util.h"

#define SYNTAX  "align_file -len min_length -pct min_ident " \
                  "[-overlaps interval_file [-inseq2] [-exons-only]] " \
                  "> output_file"

enum { BUFSIZE = 1000 };

static void get_args (int argc, char **argv, char **alignfile, int *min_len,
                      int *min_pct, char **intfile, bool *inseq2, bool *exonly);
static int get_intervals (char *intfile, int **start, int **end);
static void print_overlaps (Segment *seg, int i, int *start, int *end,
                            char **gname, int nint, bool exonly, bool inseq2);

int main (int argc, char **argv) {
	char *alignfile = NULL;
	int min_len = 0;
	int min_pct = 0;
	char *intfile = NULL;
	bool inseq2 = 0;
	bool exonly = 0;
	Segment *seg;                      /* array of alignment info */
	int size;                          /* allocated size of seg array */
	int *start = NULL, *end = NULL;    /* intervals (parallel arrays) */
	char **gname = NULL;               /* more gene/exon info */
	char **ename = NULL;
	int *dir = NULL;
	int nint;                          /* number of intervals */
	char type;
	char *hdr2 = NULL, *printed = NULL;
	int i, len, nhits = 0;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &alignfile, &min_len, &min_pct, &intfile,
	                      &inseq2, &exonly);
	if (!alignfile || min_len < 1 || min_pct < 1 || min_pct > 100)
		fatal (usage (SYNTAX));

	/* read alignment file into Segment array */
	size = get_align (alignfile, &seg);
	if (has_multi_contig (seg) && inseq2)
		fatalf ("2nd-sequence overlap intervals do not work with " \
		        "multiple contigs.");

	/* read interval file into parallel arrays */
	if (exonly)
		nint = get_exons (intfile, &start, &end, &dir, &gname, &ename, -1);
	else
		nint = get_intervals (intfile, &start, &end);

	/* print title */
	printf ("Gap-free segments in %s", alignfile);
	printf (" with length >= %d bp and identity >= %d%%", min_len, min_pct);
	if (intfile) {
		printf (",\nshowing %s from %s that overlap them in sequence %d",
		        (exonly ? "exons" : "intervals"), intfile, (inseq2 ? 2 : 1));
	}
	printf (":\n\n");

	/* find strong matches */
	for (i = 0; (type = seg[i].type) != 'z'; i++) {
		if (type == 's')
			hdr2 = seg[i].hdr2;
		else if (type == 'l' && seg[i].score >= min_pct &&
				(len = abs (seg[i].x1 - seg[i].x0) + 1) >= min_len) {
			if (!hdr2)
				fatalf ("Missing sequence header for segment %d.", i);
			if (!strsame (hdr2, printed)) {
				printf ("%s\n", hdr2);
				printed = hdr2;
			}
			printf ("    %d-%d <--> %d-%d %d%% (%d nt)",
			        seg[i].x0, seg[i].x1, seg[i].y0, seg[i].y1,
			        seg[i].score, len);
			if (seg[i].reverse)
				printf (" (reverse complement)");
			print_overlaps (seg, i, start, end, gname, nint, exonly, inseq2);
			putchar ('\n');
			nhits++;
		}
	}
	if (!nhits)
		printf ("No aligned segments matched these criteria.\n");
	else
		printf ("\n%d hit%s.\n", nhits, (nhits == 1 ? "" : "s"));

	/* clean up */
	free_align (seg, size);
	if (exonly)
		free_exons (start, end, dir, gname, ename, nint);
	else {
		free (start);
		free (end);
	}
	return 0;
}

/* Retrieve command-line arguments */
static void get_args (int argc, char **argv, char **alignfile,
                      int *min_len, int *min_pct, char **intfile,
                      bool *inseq2, bool *exonly) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-len")) {
			if (!(*min_len = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid min_length value.");   /* catch non-ints */
		}
		else if (strsame (argv[i], "-pct")) {
			if (!(*min_pct = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid min_ident value.");    /* catch non-ints */
		}
		else if (strsame (argv[i], "-overlaps"))
			*intfile = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-inseq2"))
			*inseq2 = 1;
		else if (strsame (argv[i], "-exons-only"))
			*exonly = 1;
		else if (!*alignfile)
			*alignfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* Read interval file into parallel arrays */
static int get_intervals (char *intfile, int **start, int **end) {
	FILE *fp;
	char buf [BUFSIZE];
	char head [BUFSIZE], mid [BUFSIZE], *tail;   /* non-addr parts of buf */
	int a1, a2;                                  /* interval endpoints */
	char *p;
	int i;

	if (!intfile)
		return 0;

	/* allocate parallel arrays */
	fp = ckopen (intfile, "r");
	i = 0;
	while (fgets (buf, BUFSIZE, fp)) {
		p = skip_ws (buf);
		if (*p && !starts (p, "#"))
			i++;
	}
	*start = (int *) ckalloc (i * sizeof (int));
	*end   = (int *) ckalloc (i * sizeof (int));

	/* read intervals */
	rewind (fp);
	i = 0;
	while (fgets (buf, BUFSIZE, fp)) {
		if (feat_parse_line (buf, head, &a1, mid, &a2, &tail)) {
			(*start)[i] = a1;
			(*end)[i]   = a2;
			i++;
		}
	}

	fclose (fp);
	return i;
}

/* Print overlaps from interval file */
static void print_overlaps (Segment *seg, int i, int *start, int *end,
                            char **gname, int nint, bool exonly, bool inseq2) {
	int j, count = 0;

	for (j = 0; j < nint; j++) {
		if (exonly && gname[j]) {    /* skip gene and cds */
			j++;
			continue;
		}
		if (seg_overlaps (seg, i, start[j], end[j], inseq2)) {
			printf ("%s %d-%d", (!count ? "; overlaps" : ","),
			        start[j], end[j]);
			count++;
		}
	}
}

